package org.jconfig;

import java.io.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Properties;

import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.w3c.dom.*;


public class ConfigurationManager
{
	private static Hashtable parameter;
	private static Vector categories;
	private static ConfigurationManager cfgmgr = null;
	private boolean validate = false;
	
	public ConfigurationManager()
	{	
		parameter = new Hashtable();		
		categories = new Vector();
		addCategory("general");
	}
   /**
    * This method returns the value for a particular key
	* The category is "general".
	*
	* @param key the name of the property
	* @return the value as String or null
	*/	
	public String getProperty(String key)
	{
		return getProperty(key,null,null);
	}
   /**
    * This method returns the value for a particular key.
	* If this key does not exist then the method returns
	* the defaultValue. The category is "general".
	*
	* @param key the name of the property
	* @param defaultValue the value that will be returned if the property cannot be found
	* @return the value as String or the defaultValue
	*/		
	public String getProperty(String key,String defaultValue)
	{
		return getProperty(key,defaultValue,null);
	}
   /**
    * This method returns the value for a particular key in
	* the certain category.
	* If this key does not exist then the method returns
	* the defaultValue
	*
	* @param key the name of the property
	* @param defaultValue the value that will be returned if the property cannot be found
	* @param category the name of the category 
	* @return the value as String or the defaultValue
	*/		
	public String getProperty(String key,String defaultValue,String category)
	{
		if ( key == null )
			return defaultValue;
			
		String mykey;
		if ( category != null )
			mykey = category+"/"+key;
		else
			mykey = "general/"+key;		
		if ( parameter.containsKey((String)mykey) )
		{
			return (String)parameter.get((String)mykey);
		}	
		else
			return defaultValue;
	}
   /**
    * This method returns an instance of the ConfigurationManager.
	* If there is no instance then it will create a new one.
	* In order to use the ConfigurationManager do the following:
	* private ConfigurationManager cfgmgr = ConfigurationManager.getInstance();
	*
	* @return an instance of the ConfigurationManager
	*/	
	public static ConfigurationManager getInstance()
	{
		if ( cfgmgr != null )
			return cfgmgr;
				
		cfgmgr = new ConfigurationManager();
		return cfgmgr;
	}	
   /**
    * This method returns an Enumeration of all categories.
	* The names are stored as Strings inside the Enumeration.	
	*
	* @return an enumeratiuon containing all names of all categories
	*/
	public Enumeration getCategoryNames()
	{
		return categories.elements();
	}
   /**
    * This method will return a Hashtable that contains all
	* properties for the category "general". The properties
	* are stored as key and value both as strings.
	*
	* @return a Hashtable that contains all properties for the category "general"
	*/	
	public Hashtable getProperties()
	{
		return getProperties(null);
	}
   /**
    * This method will return a Hashtable that contains all
	* properties for the selected category. The properties
	* are stored as key and value both as strings.
	*
	* @param category the name of the category
	* @return a Hashtable that contains all properties for this category
	*/		
	public Hashtable getProperties(String category)
	{
		Hashtable retValues = new Hashtable();
		// if no category was passed to us then we take "general"
		if ( category == null )
			category = "general";
		String currentKey = null;
		// get all the keys from our Hashtable
		Enumeration keys = parameter.keys();
		while ( keys.hasMoreElements() )
		{
			currentKey = (String)keys.nextElement();
			// let's see if the category name is inside the key-string
			if ( currentKey.indexOf(category+"/") != -1 )
			{
				// now extract the category name from the key and 
				// store the name and the value
				String name = currentKey.substring(currentKey.indexOf("/")+1);
				String value = (String)parameter.get((String)currentKey);
				retValues.put(name,value);
			}
		}	
		return retValues;	
	}
   /**
    * This method will set the value for a specific property in the category "general".
	* If this property does not exist then it will be created
	*
	* @param name the name of the property
	* @param value the value for this property
	*/	
	public void setProperty(String name,String value)	
	{
		setProperty(name,value,null);
	}
   /**
    * This method will set the value for a specific property in the selected category.
	* If this property does not exist then it will be created. If the category does
	* not exist it will be created.
	*
	* @param name the name of the property
	* @param value the value for this property
	* @param category the name of the category
	*/		
	public void setProperty(String name,String value,String category)
	{
		// we use the default one if it is not specified
		if ( category == null )
			category = "general";
		// it is safe to use addCategory since it will check
		// if this category already exists and return quitely	
		addCategory(category);			
		// now we save the property
		String mykey = category+"/"+name;		
		parameter.put(mykey,value);				
	}
	
	public void addCategory(String category)
	{
		if ( categories.indexOf(category) == -1 )
		{
			categories.add(category);			
		}
	}
	
   /**
    * This method will save the current categories and their properties
	* to a file
	*
	* @param file the file that will be generated
	*/		
	public void store(File file) throws ConfigurationManagerException
	{
		// this will store the properties for a certain category
		Hashtable props = new Hashtable();
		try
		{
			// let's open the file
			FileWriter fw = new FileWriter(file);
			// and write the header
			fw.write("<?xml version=\"1.0\" ?>\n");
			fw.write("<properties>\n");
			// now we walk through all categories
			for ( int i =0; i < categories.size(); i++)
			{
				fw.write("  <category name='"+(String)categories.get(i)+"'>\n");
				// now we get all properties for this category
				props = getProperties((String)categories.get(i));
				Enumeration lprops = props.keys();
				// here we walk through the properties
				while ( lprops.hasMoreElements() )
				{
					String currentKey = (String)lprops.nextElement();
					// and write it to the file
					fw.write("    <property name='"+currentKey+"' value='"+(String)props.get(currentKey)+"'/>\n");
				}
				fw.write("  </category>\n");			
			}
			fw.write("</properties>\n");
			// close the file because we are done
			fw.close();
		}
		catch (Exception e)
		{
			throw new ConfigurationManagerException("The file cannot be saved");
		}
	}
	
	private void processProperties(Document doc)
	{
		String currentCategory;
		
		// first we get all nodes where the element is category
		NodeList nl = doc.getElementsByTagName("category");
		for (int i = 0 ; i< nl.getLength(); i++)
		{
			// now we get every node from the list
			Node n = nl.item(i);
			// and get the name attribute for this category
			NamedNodeMap curAtt=n.getAttributes();
			Node curNode = curAtt.getNamedItem("name");
			currentCategory = curNode.getNodeValue();				
			// now we process all children for this category								
			for (Node child = n.getFirstChild(); child != null; child = child.getNextSibling())
		    {
				if ( child.getNodeName().equals("property") )
				{
					// we have found a property element and now we grab the name and value
					// attributes
					NamedNodeMap myAtt=child.getAttributes();
					Node myNode = myAtt.getNamedItem("name");
					String name = myNode.getNodeValue();
					myNode=myAtt.getNamedItem("value");
					String value=myNode.getNodeValue();
					// if we have both then lets store it
					if ( name != null && value != null )
					{
						// the key is always category/name
						// e.g. general/congig_file
						setProperty(name,value,currentCategory);
					}
				}
        	}
		}
	}
   /**
    * This method will set the flag if validation is used for
	* parsing the file/stream from URL or not
	*
	* @param validate true if the content should be validated first
	*
	*/	
	public void setValidation(boolean validate)
	{
		this.validate = validate;
	}
   /**
    * This method will read in a file and generate the properties
	*
	* @param a file
	* @throws ConfigurationManagerException if the file cannot be processed
	*/	
	public synchronized void load(File file) throws ConfigurationManagerException
	{
		cfgmgr = new ConfigurationManager();
		
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		// set the validation flag (true if it should be validated)
        dbf.setValidating(validate);
        DocumentBuilder db = null;
        try 
		{
            db = dbf.newDocumentBuilder();
        } 
		catch (ParserConfigurationException pce) 
		{
            throw new ConfigurationManagerException("The parser cannot create a new document builder: "+pce.getMessage());
        }
		// if we have to validate then we need to define an error handler here
  		if ( validate )
		{
			try
			{
				// Set an ErrorHandler before parsing
		        OutputStreamWriter errorWriter =
		            new OutputStreamWriter(System.err, "UTF-8");
		        db.setErrorHandler(
		            new MyErrorHandler(new PrintWriter(errorWriter, true)));
			}
			catch (Exception e)
			{
				throw new ConfigurationManagerException("The parser cannot set up the error handler");
			}
		}
		Document doc = null;
        try 
		{
            doc = db.parse(file);
        } 
		catch (SAXException se) 
		{			
            throw new ConfigurationManagerException("The parser cannot parse the file: "+se.getMessage());
        } 
		catch (IOException ioe) 
		{		
            throw new ConfigurationManagerException("The parser cannot open the file: "+ioe.getMessage());
        }
		processProperties(doc);
		
	}
		
   /**
    * This method will read the content from an URL and generate the properties
	*
	* @param URL the URL as string
	* @throws ConfigurationManagerException if the file cannot be processed
	*/	
	public synchronized void load(String theURL) throws ConfigurationManagerException
	{
		cfgmgr = new ConfigurationManager();
		
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        dbf.setValidating(validate);
        DocumentBuilder db = null;
        try 
		{
            db = dbf.newDocumentBuilder();
        } 
		catch (ParserConfigurationException pce) 
		{
            throw new ConfigurationManagerException("The parser cannot create a new document builder: "+pce.getMessage());
        }
  		// if we have to validate then we need to define an error handler here
  		if ( validate )
		{
			try
			{
				// Set an ErrorHandler before parsing
		        OutputStreamWriter errorWriter =
		            new OutputStreamWriter(System.err, "UTF-8");
		        db.setErrorHandler(
		            new MyErrorHandler(new PrintWriter(errorWriter, true)));
			}
			catch (Exception e)
			{
				throw new ConfigurationManagerException("The parser cannot set up the error handler");
			}
		}
        Document doc = null;
        try 
		{
            doc = db.parse(theURL);
        } 
		catch (SAXException se) 
		{			
            throw new ConfigurationManagerException("The parser cannot parse the XML: "+se.getMessage());
        } 
		catch (IOException ioe) 
		{		
            throw new ConfigurationManagerException("The parser cannot open the file: "+ioe.getMessage());
        }
		processProperties(doc);
		
	}
   /**
    * This method will store all categories and properties to a Java-properties file.
	* All properties will be in the category "general". It will override an 
	* existing configuration.
	*
	* @param file the file
	* @throws ConfigurationManagerException if the file cannot be processed
	*/		
	public void loadProperties(File file) throws ConfigurationManagerException
	{
		try
		{
			FileInputStream input = new FileInputStream(file);
			Properties props = new Properties();
			props.load(input);
			Enumeration propertyNames = props.propertyNames();
			cfgmgr = new ConfigurationManager();
			while ( propertyNames.hasMoreElements() )
			{
				String name = (String)propertyNames.nextElement();
				String value = props.getProperty(name);
				setProperty(name,value);
			}				
		}
		catch (Exception e)
		{
			throw new ConfigurationManagerException("The file cannot be loaded");
		}
	}
   /**
    * This method will generate a Java-properties file.
	* All categories will be saved as comments and only the properties are written to the file.
	* The output is category.property=value
	*
	* @param file the file
	* @throws ConfigurationManagerException if the file cannot be processed
	*/			
	public void storeProperties(File file) throws ConfigurationManagerException
	{
		// this will store the properties for a certain category
		Hashtable props = new Hashtable();
		String currentCategory = new String();
		try
		{
			// let's open the file
			FileWriter fw = new FileWriter(file);
			// and write the header
			fw.write("#\n");
			fw.write("# automatically generated properties file\n");
			fw.write("#\n");
			// now we walk through all categories
			for ( int i =0; i < categories.size(); i++)
			{
				currentCategory = (String)categories.get(i);
				// write the category as comment
				fw.write("#\n");
				fw.write("# category: " + currentCategory+"\n");
				fw.write("#\n");
				// now we get all properties for this category
				props = getProperties((String)categories.get(i));
				Enumeration lprops = props.keys();
				// here we walk through the properties
				while ( lprops.hasMoreElements() )
				{
					String currentKey = (String)lprops.nextElement();
					// and write it to the file
					// form category.property=value
					fw.write(currentCategory+"."+currentKey+"="+(String)props.get(currentKey)+"\n");
				}
			}
			// close the file because we are done
			fw.close();
		}
		catch (Exception e)
		{
			throw new ConfigurationManagerException("The file cannot be saved");
		}
	}
   /**
    * This method will write a Java-properties file and fill in the properties for the
	* particular category. The category will be saved as comment and 
	* only the properties are written to the file. The output is property=value.
	* The category is not included in the property name.
	*
	* @param file the file
	* @throws ConfigurationManagerException if the file cannot be processed
	*/				
	public void storeProperties(File file,String category) throws ConfigurationManagerException
	{
		// this will store the properties for a certain category
		Hashtable props = new Hashtable();
		try
		{
			// let's open the file
			FileWriter fw = new FileWriter(file);
			// and write the header
			fw.write("#\n");
			fw.write("# automatically generated properties file");
			fw.write("#\n");
			fw.write("#\n");
			fw.write("# category: " + category+"\n");
			fw.write("#\n");
			// now we get all properties for this category
			props = getProperties(category);
			Enumeration lprops = props.keys();
			// here we walk through the properties
			while ( lprops.hasMoreElements() )
			{
				String currentKey = (String)lprops.nextElement();
				// and write it to the file
				// form category.property=value
				fw.write(currentKey+"="+(String)props.get(currentKey)+"\n");
			}			
			// close the file because we are done
			fw.close();
		}
		catch (Exception e)
		{
			throw new ConfigurationManagerException("The file cannot be saved");
		}
	}
	
	//
	// This is an inner class
	//
	// Error handler to report errors and warnings
	//
    private static class MyErrorHandler implements ErrorHandler 
	{
    
        private PrintWriter out;

        MyErrorHandler(PrintWriter out) 
		{
            this.out = out;
        }

        /**
         * Returns a string describing parse exception details
         */
        private String getParseExceptionInfo(SAXParseException spe) 
		{
            String systemId = spe.getSystemId();
            if (systemId == null) 
			{
                systemId = "null";
            }
            String info = "URI=" + systemId +
                " Line=" + spe.getLineNumber() +
                ": " + spe.getMessage();
            return info;
        }

        // The following methods are standard SAX ErrorHandler methods.
        // See SAX documentation for more info.

        public void warning(SAXParseException spe) throws SAXException 
		{
            out.println("Warning: " + getParseExceptionInfo(spe));
        }
        
        public void error(SAXParseException spe) throws SAXException 
		{
            String message = "Error: " + getParseExceptionInfo(spe);
            throw new SAXException(message);
        }

        public void fatalError(SAXParseException spe) throws SAXException 
		{
            String message = "Fatal Error: " + getParseExceptionInfo(spe);
            throw new SAXException(message);
        }
    }	
	
}
